unit BSP;

{ BSPTree (c)2K by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
 SysUtils;

Type
 PWall =^TWall;
 TWall =record
 // .map
  x1,y1,x2,y2 :integer;
  h1,h2,h3,h4 :integer;
  Texture     :integer;
  FrontFloor  :integer;
  FrontCeiling:integer;
  BackFloor   :integer;
  BackCeiling :integer;
  TextureX    :integer;
  TextureY    :integer;
 // .bsp
  LeftWall    :integer;
  RightWall   :integer;
 // linked list of walls...
  Next        :PWall;
 end;

var
// first line of both MAP and BSP file
 wallcount,playerx,playery,playerz,playera:integer;

 Walls:PWall=nil;
 Nodes:PWall=nil;
 Tree :PWall=nil;
 NodeCount:integer=0;
 Root:integer;

function LoadMap(AFileName:string):integer;
procedure SaveBSP(AFileName:string);

implementation

Function NewWall:PWall; // Create a new wall, automaticaly add it to Walls list
 begin
  New(Result);
  Result^.Next:=Walls;
  Walls:=Result;
 end;

Function AddNode(AWall:PWall):integer; // Build list of nodes, return node index
 begin
  AWall^.Next:=nil;
  if Tree=nil then begin
   NodeCount:=1;
   Tree:=AWall;
  end else begin
   inc(NodeCount);
   Nodes^.Next:=AWall;
  end;
  Nodes:=AWall;
  Result:=NodeCount-1;
 end;

Function Side(w:PWall;x,y:single):single; // witch side of line (x,y)
 begin
  with w^ do result:=(y1-y2)*(x1-x)-(x1-x2)*(y1-y);
 end;

Function Split(w:PWall; a,b:single):PWall;
 var
  x,y:integer;
 begin
  inc(WallCount);
 // duplicate wall
  New(Result); Move(w^,Result^,SizeOf(TWall));
 // compute intersection
  x:=round((w^.x2*a+w^.x1*b)/(a+b));
  y:=round((w^.y2*a+w^.y1*b)/(a+b));
 // First wall
  Result^.x1:=x;
  Result^.y1:=y;
 // Second wall
  w^.x2:=x;
  w^.y2:=y;
 end;

Function GetBest(Var Walls:PWall):PWall;
 var
  i:PWall;
  j:PWall;
  k,l:integer;
  m,ml:integer;
  v1,v2:single;
 begin
  m:=32767;
  ml:=32767;
  i:=Walls;
  while i<>nil do begin // for each wall of the list
   k:=0;
   l:=0;
   j:=Walls;
   while j<>nil do begin // for each wall of the list
    if (i<>j) then begin  // except the wall himself :)
     v1:=Side(i,j^.x1,j^.y1); // witch side
     v2:=Side(i,j^.x2,j^.y2);
     if (v1<=0)and(v2<=0) then inc(l) else
     if (v1>0)and(v2>0) then dec(l) else begin
      inc(k);
      inc(l);
     end;
    end;
    j:=j^.next; // next wall
   end;
   l:=abs(l);
   if (l<ml) or ((l=ml)and(k<m)) then begin // best balance
    m:=k;
    ml:=l;
    Result:=i;
   end;
   i:=i^.next; // next wall
  end;

  // remove the 'best wall' from the list
  if Result=Walls then
   Walls:=Result^.Next
  else begin
   i:=Walls;
   while (i^.next<>nil) do begin
    if i^.next=result then begin
     i^.next:=Result^.next;
     exit;
    end;
    i:=i^.next;
   end;
  end;
 end;

Function BuildTree(Wall:PWall):integer;
 var
  Best:PWall;
  n:PWall;
  v1,v2:single;
  l1,l2:PWall;
 begin

  if Wall=nil then begin
   Result:=0;
   exit;
  end;

  Best:=GetBest(Wall);

  l1:=nil; // right list
  l2:=nil; // left list

  while Wall<>nil do begin
   n:=Wall^.Next;

   v1:=Side(Best,Wall^.x1,Wall^.y1);
   v2:=Side(Best,Wall^.x2,Wall^.y2);

   if (v1<=0) and (v2<=0) then begin
    Wall^.Next:=l1; l1:=Wall; // add wall to Right list
   end else
   if (v1>=0) and (v2>=0) then begin
    Wall^.Next:=l2; l2:=Wall; // add wall to Left list
   end else
   if (v1<0) then begin
    Wall^.Next:=l1; l1:=Wall; // add partial wall to Right list
    Wall:=Split(Wall,-v1,v2); // cut wall
    Wall^.Next:=l2; l2:=Wall; // add other part to Left list
   end else begin
    Wall^.next:=l2; l2:=Wall; // add partial wall to Left list
    Wall:=split(Wall,v1,-v2); // cut wall
    Wall^.next:=l1; l1:=Wall; // add other part to Right list
   end;

   Wall:=n; // next wall in the list
  end;

  Result:=AddNode(Best);
  Best^.RightWall:=BuildTree(l2);
  Best^.LeftWall:=BuildTree(l1);
 end;

Procedure Clear;
 begin
  while Tree<>nil do begin
   Walls:=Tree.Next;
   Dispose(Tree);
   Tree:=Walls;
  end;
  Walls:=nil;
  Nodes:=nil;
  NodeCount:=0;
 end;

function LoadMap(AFileName:string):integer;
 var
  map:textfile;
  i:integer;
 begin
  Clear;

  AssignFile(map,AFileName);
  Reset(map);
  if IOResult<>0 then Raise Exception.Create('source map not found');
  readln(map,wallcount,playerx,playery,playerz,playera);
  result:=wallcount;
  for i:=0 to wallcount-1 do
   with NewWall^ do
     ReadLn(map,x1,y1,x2,y2,h1,h2,h3,h4,
                Texture,FrontFloor,FrontCeiling,BackFloor,BackCeiling,
                TextureX,TextureY);
  CloseFile(map);
  Root:=BuildTree(Walls);
 end;

Procedure SaveBSP(AFileName:string);
 var
  bsp:textfile;
  Wall:PWall;
 begin
  AssignFile(bsp,AFileName);
  rewrite(bsp);
  if ioresult<>0 then raise Exception.Create('unable to create target bsp file');
  writeln(bsp,wallcount,' ',playerx,' ',playery,' ',playerz,' ',playera);

  Wall:=Tree;
  while Wall<>nil do begin
   with Wall^ do
    WriteLn(bsp,x1:6,y1:6,x2:6,y2:6,h1:5,h2:5,h3:5,h4:5,
               Texture:2,FrontFloor:2,FrontCeiling:2,BackFloor:2,BackCeiling:2,
               TextureX:4,TextureY:4,
               ' ',LeftWall:3,' ',RightWall:3);
   Wall:=Wall^.Next;
  end;

  CloseFile(bsp);
 end;

end.
