unit BSPTree1;

{ BSPTree (c)2K by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}

{ MAP files from :
   DOOM levels player by Ivaylo Beltchev
   email: ivob@geocities.com
   url:   http://www.geocities.com/SiliconValley/Bay/2234
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, BSP;

type
  float = extended;

  TForm1 = class(TForm)
    Panel1: TPanel;
    Image1: TImage;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    Load: TPanel;
    Save: TPanel;
    Quit: TPanel;
    Status: TPanel;
    View: TPanel;
    ShowTree: TPanel;
    Map: TPanel;
    procedure LoadClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SaveClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure LoadMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure LoadMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure LoadMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure QuitClick(Sender: TObject);
    procedure ShowTreeClick(Sender: TObject);
    procedure ViewClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure MapClick(Sender: TObject);
  private
    { Dclarations prives }
    Walls:TList;
    Depth:integer;
    cs,sn:float;
    fView:(vNone,vTree,vMap,v3D);
    ww,hh:integer;
    function Wall(Index:integer):PWall;
    procedure WalkTree(i:integer);
    procedure DrawWall(i:integer);
  public
    { Dclarations publiques }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
begin
 Walls:=TList.Create;
 fView:=vNone;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 Walls.Free;
end;

function TForm1.Wall(Index:integer):PWall;
 begin
  Result:=Walls[Index];
 end;

procedure TForm1.LoadClick(Sender: TObject);
var
 W:PWall;

  function TreeDepth(W:PWall):integer;
   var
    L,R:integer;
   begin
    L:=W.LeftWall;
    R:=W.RightWall;
    if L>0 then L:=TreeDepth(Wall(L));
    if R>0 then R:=TreeDepth(Wall(R));
    if R>L then L:=R;
    Result:=L+1;
   end;

begin
 if OpenDlg.Execute then begin
  Status.Caption:=IntToStr(LoadMap(OpenDlg.FileName))+' walls, '+IntToStr(NodeCount)+' nodes';
  Walls.Clear;
  W:=BSP.Tree;
  Repeat
   Walls.Add(W);
   W:=W.Next;
  until W=nil;
  Depth:=TreeDepth(Tree);
 end;
end;

procedure TForm1.SaveClick(Sender: TObject);
begin
 if SaveDlg.Execute then SaveBSP(SaveDlg.FileName);
end;

procedure TForm1.FormResize(Sender: TObject);
begin
 Image1.Picture:=nil;
 ww:=Width div 2;
 hh:=Height div 2;
 case fView of
  vTree:ShowTreeClick(Self);
  vMap :MapClick(Self);
  v3D  :ViewClick(Self);
 end;
end;

procedure TForm1.LoadMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
 with TPanel(Sender) do begin
  case Tag of
   0: begin  // mouse is entering on the component
       BevelInner:=bvRaised;
       SetCaptureControl(TPanel(Sender));
       Tag:=1;
      end;
   1: begin // mouse is moving while the Control has captured Mouse Events
       if (X<0)or(Y<0)or(X>Width)or(Y>Height) then begin
        BevelInner:=bvNone;
        SetCaptureControl(nil);
        Tag:=0;
       end else BevelInner:=bvRaised;
      end;
   2: begin // mouse is moving while the Control has captured Mouse Events, after a click
       if (X<0)or(Y<0)or(X>Width)or(Y>Height) then begin BevelInner:=bvNone; Tag:=3; end;
      end;
   3: begin
       if (X<0)or(Y<0)or(X>Width)or(Y>Height) then else begin BevelInner:=bvLowered; Tag:=2 end;
      end;
  end;
 end;
end;

procedure TForm1.LoadMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 with TPanel(Sender) do begin
  Tag:=2;
  BevelInner:=bvLowered;
  SetCaptureControl(TPanel(Sender));
 end;
end;

procedure TForm1.LoadMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 with TPanel(Sender) do begin
  Tag:=0;
  BevelInner:=bvNone;
  SetCaptureControl(nil);
 end;
end;

procedure TForm1.QuitClick(Sender: TObject);
begin
 Application.Terminate;
end;

procedure TForm1.ShowTreeClick(Sender: TObject);
 var
  h:integer;

  procedure Walk(x,y,w:integer;T:PWall);
   begin
    with Image1,Canvas do begin
     LineTo(x,y);
     Ellipse(x-2,y-2,x+2,y+2);
     w:=w div 2;
     if T.LeftWall>0 then begin
      MoveTo(x,y);
      Walk(x-w div 2,y+h,w,Wall(T.LeftWall));
     end;
     if T.RightWall>0 then begin
      MoveTo(x,y);
      Walk(x+w div 2,y+h,w,Wall(T.RightWall));
     end;
    end;
   end;

 begin
  fView:=vTree;
  Image1.Picture:=nil;
  if Walls.Count=0 then exit;
  with Image1,Canvas do begin
   h:=Height div Depth;
   MoveTo(Width div 2,1);
   Walk(Width div 2,h div 2,Width,Wall(0));
  end;
 end;

procedure TForm1.ViewClick(Sender: TObject);
begin
 fView:=v3D;
// Image1.Picture:=nil;
 if Walls.Count=0 then exit;
 cs:=cos(playera*2*PI/128);
 sn:=sin(playera*2*PI/128);
 WalkTree(0);
 Image1.RePaint;
end;

procedure TForm1.WalkTree(i:integer);
 begin
  with Wall(i)^ do begin
  { infront }
   if longint(y1-y2)*(x1-playerx)<longint(x1-x2)*(y1-playery) then begin
    if RightWall>0 then WalkTree(RightWall);
    DrawWall(i);
    if LeftWall>0 then WalkTree(LeftWall);
   end else begin
    if LeftWall>0 then WalkTree(LeftWall);
    DrawWall(i);
    if RightWall>0 then WalkTree(RightWall);
   end;
  end;
 end;

procedure TForm1.DrawWall(I:integer);
 const
  foc=256;
 var
  cl:integer;
  wx1,wy1:float;
  wx2,wy2:float;
  nx,ny,nl:float;
  wt1,wc1,wf1,wb1:float;
  wt2,wc2,wf2,wb2:float;
  wx:float;
  b:float;
  dx,dy:float;
  x1,x2:float;

  function iPoint(x,y:float):TPoint;
   begin
    result.X:=Round(x);
    result.y:=hh-Round(y);
   end;

 begin
  { origine }
  with wall(i)^ do begin
   wx1:=x1-playerx;
   wy1:=y1-playery;

   wx2:=x2-playerx;
   wy2:=y2-playery;

   wt2:=h4-playerz;
   wc2:=h3-playerz;
   wf2:=h2-playerz;
   wb2:=h1-playerz;

  end;
 { rotate wall }
  wx :=wx1*cs-wy1*sn;
  wy1:=wx1*sn+wy1*cs;
  wx1:=wx;

  wx :=wx2*cs-wy2*sn;
  wy2:=wx2*sn+wy2*cs;
  wx2:=wx;

  if (wy1<1) and (wy2<1) then exit; { unvisible }

  nx:=wx2-wx1;
  ny:=wy2-wy1;
  nl:=sqrt(nx*nx+ny*ny);
  if nl<>0 then begin
   nx:=nx/nl;
//   ny:=ny/nl;
  end;

 { project }
  if (wy1<1) then begin
   wx1:=wx2+(wx1-wx2)*(wy2-1)/(wy2-wy1);
   wy1:=1;
  end;

  if (wy2<1) then begin
   wx2:=wx1+(wx2-wx1)*(wy1-1)/(wy1-wy2);
   wy2:=1;
  end;

  x1:=(wx1*foc/wy1)+ww;
  x2:=(wx2*foc/wy2)+ww;
  if (x1>=x2) then exit; { wrong face }

  if (x2<0)or(x1>2*ww) then exit;
(*
  if (x1<0) then begin
   wy1:=wy1+(wy2-wy1)*(x2)/(x2-x1);
   x1:=0;
  end;
  if (x2>2*ww) then begin
   wy2:=wy2+(wy2-wy1)*(2*ww-x1)/(x2-x1);
   x2:=2*ww;
  end;
*)
  wt1:=wt2*foc/wy1;
  wc1:=wc2*foc/wy1;
  wf1:=wf2*foc/wy1;
  wb1:=wb2*foc/wy1;

  wt2:=wt2*foc/wy2;
  wc2:=wc2*foc/wy2;
  wf2:=wf2*foc/wy2;
  wb2:=wb2*foc/wy2;

  with Image1,Canvas do begin
   Pen.Color:=clAqua;
   Brush.Color:=clAqua;
   if (wt1<hh)and(wt2<hh) then
    Polygon([iPoint(x1,hh),iPoint(x2,hh),iPoint(x2,wt2),iPoint(x1,wt1)]);
   Pen.Color:=clBlue;
   Brush.Color:=clBlue;
   Polygon([iPoint(x1,wb1),iPoint(x2,wb2),iPoint(x2,hh-Height),iPoint(x1,hh-Height)]);
   cl:=Round(128*abs(nx)+64);
   Pen.Color:=RGB(cl,cl,cl);
   Brush.Color:=Pen.Color;
   if (wt1>wc1)and(wt2>wc2) then begin
    Polygon([iPoint(x1,wt1),iPoint(x2,wt2),iPoint(x2,wc2),iPoint(x1,wc1)]);
   end;
   if (wf1>wb1)and(wf2>wb2) then begin
    Polygon([iPoint(x1,wf1),iPoint(x2,wf2),iPoint(x2,wb2),iPoint(x1,wb1)]);
   end;
  end;
 end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
const
 speed=5;
begin
 if fView<>v3D then exit;
 case Key of
  vk_left  : dec(PlayerA);
  vk_Right : inc(PlayerA);
  vk_up    : begin
   Inc(PlayerX,Round(Speed*sn));
   Inc(PlayerY,Round(Speed*cs));
  end;
  vk_down  : begin
   Dec(PlayerX,Round(Speed*sn));
   Dec(PlayerY,Round(Speed*cs));
  end;
  vk_next  : dec(playerz);
  vk_prior : inc(playerz);
  else exit;
 end;
 ViewClick(Self);
end;

procedure TForm1.MapClick(Sender: TObject);
var
 i:integer;
 scale,s1:single;
 xmin,ymin,xmax,ymax:integer;
 ox,oy:integer;
begin
 fView:=vMap;
 Image1.Picture:=nil;
 if Walls.Count=0 then exit;
 with Wall(0)^ do begin
  if x1<x2 then begin xmin:=x1; xmax:=x2 end else begin xmin:=x2; xmax:=x1; end;
  if y1<y2 then begin ymin:=y1; ymax:=y2 end else begin ymin:=y2; ymax:=y1; end;
 end;
 for i:=1 to Walls.Count-1 do begin
  with Wall(i)^ do begin
   if x1<xmin then xmin:=x1;
   if x2<xmin then xmin:=x2;
   if x1>xmax then xmax:=x1;
   if x2>xmax then xmax:=x2;
   if y1<ymin then ymin:=y1;
   if y2<ymin then ymin:=y2;
   if y1>ymax then ymax:=y1;
   if y2>ymax then ymax:=y2;
  end;
 end;

 s1:=(Image1.Width-10)/(xmax-xmin);
 scale:=(Image1.Height-10)/(ymax-ymin);
 if scale>s1 then scale:=s1;

 ox:=(Image1.Width-Round(scale*(xmax-xmin))) div 2;
 oy:=Image1.Height-(Image1.Height-Round(scale*(ymax-ymin))) div 2;

 with Image1,Canvas do begin
  Pen.Color:=clBlack;
  for i:=0 to Walls.Count-1 do begin
   with Wall(i)^ do begin
    MoveTo(ox+Round(scale*(x1-xmin)),oy-Round(scale*(y1-ymin)));
    LineTo(ox+Round(scale*(x2-xmin)),oy-Round(scale*(y2-ymin)));
   end;
  end;
  Pen.Color:=clRed;
  ox:=ox+Round(scale*(PlayerX-xmin));
  oy:=oy-Round(scale*(PlayerY-ymin));
  Ellipse(ox-3,oy-3,ox+3,oy+3);
 end;
end;

end.
